#include <Adafruit_NeoPixel.h>
 

#define CONTROL        0
#define BLOCKLY        1


#define ALIVE          0
#define VIBE           1


#define GET            1
#define SET            2
#define RESET          3


#define PIN_LEFT_MOTOR_FW      2
#define PIN_LEFT_MOTOR_BW      4
#define PIN_RIGHT_MOTOR_FW     8
#define PIN_RIGHT_MOTOR_BW     7


#define PIN_LED               11


#define PIN_BUZZER             3

#define PIN_VIBE              A0


#define PIN_GEAR_MOTOR_FW      9
#define PIN_GEAR_MOTOR_BW     10


#define PIN_ANGLE_MOTOR_FW     6
#define PIN_ANGLE_MOTOR_BW     5


#define PIN_NEOPIXEL          12
#define NEOPIXEL_NUM          36

Adafruit_NeoPixel pixels(NEOPIXEL_NUM, PIN_NEOPIXEL, NEO_GRB + NEO_KHZ800);


int mode             = 0;
int life             = 3;
int shootMotorState  = 0;


int GEAR_MAX_SPEED = 200;
int GEAR_MIN_SPEED = 0;
int GEAR_CATCH_SPEED = 80;


int ANGLE_MAX_SPEED = 220;
int ANGLE_MIN_SPEED = 0;

union {
  byte byteVal[4];
  float floatVal;
  long longVal;
} val;

union {
  byte byteVal[2];
  short shortVal;
} valShort;

const int START_VAL = 255;
const int DATA_LEN = 14;
const int END_VALUE = 10;

int beforeBuzNote = 0;
int rotation = 1000;

unsigned char buf[DATA_LEN];
unsigned char prevc=0;
int index = 0;
boolean isStart = false;

int vibeState = 0;
boolean isReload = false;


void pixelAllCtrl(int rv, int gv, int bv) {
  for (int num = 0; num < NEOPIXEL_NUM; num++) {
    pixels.setPixelColor(num, pixels.Color(rv, gv, bv));
  }
  pixels.show();
}

void initPorts() {
  for (int i=2; i < 14; i++) {
    pinMode(i, OUTPUT);
    digitalWrite(i, LOW);
  }
  // 각도 조절 모터 - Cannon
  pinMode(PIN_ANGLE_MOTOR_FW, OUTPUT);
  pinMode(PIN_ANGLE_MOTOR_BW, OUTPUT);
  digitalWrite(PIN_ANGLE_MOTOR_FW, LOW);
  digitalWrite(PIN_ANGLE_MOTOR_BW, LOW);
}

void updateDigitalPort(unsigned char c) {
  if (c == DATA_LEN && isStart == false) {
    if (prevc == START_VAL) {
      isStart = true;
      index = 1;
      writeBuffer(index-1, prevc);
      writeBuffer(index, c);
    }
  } else {
    prevc = c;
    if (isStart) {
      writeBuffer(index, c);
    }
  }

  index++;
  
  if (index > DATA_LEN) {
    isStart = false;
    index = 0;
  }

  if (isStart && index == DATA_LEN && index > 3) {
    isStart = false;
    parseData();
    index = 0;
  }
}

boolean isCheckOk () {
  if (buf[DATA_LEN-1] == END_VALUE) {
    int checkSum = buf[DATA_LEN-2];
    int sumValue = 0;
    for (int i=2; i < DATA_LEN-2; i++) {
      sumValue += buf[i];
    }
    sumValue = sumValue & 0xff;
    if (checkSum == sumValue) {
      return true;
    }
  }
  return false;
}

unsigned char readBuffer(int index) {
  return buf[index];
}

void buzCtrl(int buzOctave, int buzNote) {
  int note = (buzOctave << 8) + buzNote;
  if (beforeBuzNote != note) {
    beforeBuzNote = note;
    if (note > 30) {
      tone(PIN_BUZZER, note);
    } else {
      noTone(PIN_BUZZER);
    }
  }
}

void motorCtrl(int motorState) {
  switch(motorState) {
    case 0: //Stop
      digitalWrite(PIN_RIGHT_MOTOR_FW, LOW);
      digitalWrite(PIN_RIGHT_MOTOR_BW, LOW);
      digitalWrite(PIN_LEFT_MOTOR_FW, LOW);
      digitalWrite(PIN_LEFT_MOTOR_BW, LOW);
      break;
    case 1: // Forward
      digitalWrite(PIN_RIGHT_MOTOR_FW, HIGH);
      digitalWrite(PIN_RIGHT_MOTOR_BW, LOW);
      digitalWrite(PIN_LEFT_MOTOR_FW, HIGH);
      digitalWrite(PIN_LEFT_MOTOR_BW, LOW);
      break;
    case 2: // Turn Left
      digitalWrite(PIN_RIGHT_MOTOR_FW, HIGH);
      digitalWrite(PIN_RIGHT_MOTOR_BW, LOW);
      digitalWrite(PIN_LEFT_MOTOR_FW, LOW);
      digitalWrite(PIN_LEFT_MOTOR_BW, HIGH);
      break;
    case 3: // Turn Right
      digitalWrite(PIN_RIGHT_MOTOR_FW, LOW);
      digitalWrite(PIN_RIGHT_MOTOR_BW, HIGH);
      digitalWrite(PIN_LEFT_MOTOR_FW, HIGH);
      digitalWrite(PIN_LEFT_MOTOR_BW, LOW);
      break;
    case 4: // Backward
      digitalWrite(PIN_RIGHT_MOTOR_FW, LOW);
      digitalWrite(PIN_RIGHT_MOTOR_BW, HIGH);
      digitalWrite(PIN_LEFT_MOTOR_FW, LOW);
      digitalWrite(PIN_LEFT_MOTOR_BW, HIGH);
      break;
  }
}

void ledCtrl(int ledPin, int ledValue) {
  if (ledValue == 255) {
    digitalWrite(ledPin, HIGH);
  } else if (ledValue == 0) {
    digitalWrite(ledPin, LOW);
  } else {
    analogWrite(ledPin, ledValue);
  }
}

void controlShootMotorCtrl() {
  
  if (shootMotorState == 2) {
    
    shootMotorCtrl(shootMotorState);
    delay(2000);
  }
  
  if (shootMotorState == 0 && life == 3) {
    full_energy();
}
  
  while (shootMotorState == 2 && life > 0) {
  
    int vibe_value =  analogRead(PIN_VIBE);
    if (vibe_value > 600) {
   
      pixels.clear();
      theaterChase(pixels.Color(40, 40, 40), 10);
      writeValue(0xff);
      writeValue(0x7);
      writeValue(VIBE);
      writeValue(1); 
      writeValue(--life);
      int checkSum = 0;
      checkSum = (VIBE + 1 + life) & 0xff;
      writeValue(checkSum);
      writeEnd();
       

      for(int i=0; i < 2; i++) {
        tone(PIN_BUZZER, 194, 100);
        delay(150);
        tone(PIN_BUZZER, 212, 100);
        delay(150);
      }

      delay(400);

      if (life == 0) {

          colorWipe(pixels.Color(40, 0, 40), 10); // Red
          tone(PIN_BUZZER, 1046, 80);
          colorWipe(pixels.Color(0, 40, 40), 10); // Green
          tone(PIN_BUZZER, 784, 80);
          colorWipe(pixels.Color(0, 0, 0), 10); // Blue
          tone(PIN_BUZZER, 659, 80);  
        //colorWipe(pixels.Color(0, 0, 0, 255), 30); // White RGBW
          theaterChase(pixels.Color(40, 40, 0), 10); // White
          tone(PIN_BUZZER, 523, 1000);
          delay(1000);
          pixels.clear();
          pixels.show();

           
          for(int i=0; i < 10; i++) {
          tone(PIN_BUZZER, 33, 300);
          delay(350);
     
        }
        life -= 1;
      } else if (life == 3) {
        full_energy(); 
      } else if (life == 2) {
        twostep_energy();
      } else if (life == 1) {
        onestep_energy();
      }
    }
    
    bluetooth_read();
  }
  
  if (shootMotorState != 2) {
    shootMotorCtrl(shootMotorState);
  }
}

void shootMotorCtrl(int shootMotorState) {
  if (shootMotorState == 0) { // 정지
    analogWrite(PIN_GEAR_MOTOR_FW, GEAR_MIN_SPEED);
    analogWrite(PIN_GEAR_MOTOR_BW, GEAR_MIN_SPEED);
  } else if (shootMotorState == 1) { 
    isReload = true;
    analogWrite(PIN_GEAR_MOTOR_FW, GEAR_MAX_SPEED);
    analogWrite(PIN_GEAR_MOTOR_BW, GEAR_MIN_SPEED);
  } else if (shootMotorState == 2) { 
    if (isReload) {
      analogWrite(PIN_GEAR_MOTOR_FW, GEAR_MIN_SPEED);
      analogWrite(PIN_GEAR_MOTOR_BW, GEAR_MAX_SPEED);
      delay(80);
      isReload = false;
    }
    analogWrite(PIN_GEAR_MOTOR_FW, GEAR_MIN_SPEED);
    analogWrite(PIN_GEAR_MOTOR_BW, GEAR_MIN_SPEED);
  } else if (shootMotorState == 3) { 
    analogWrite(PIN_GEAR_MOTOR_FW, GEAR_CATCH_SPEED);
    analogWrite(PIN_GEAR_MOTOR_BW, GEAR_MIN_SPEED);
  }
}

void angleMotorCtrl(int angleMotorState) {
  if (angleMotorState == 1) { 
    analogWrite(PIN_ANGLE_MOTOR_FW, ANGLE_MAX_SPEED);
    analogWrite(PIN_ANGLE_MOTOR_BW, ANGLE_MIN_SPEED);
  } else if (angleMotorState == 2) { 
    analogWrite(PIN_ANGLE_MOTOR_FW, ANGLE_MIN_SPEED);
    analogWrite(PIN_ANGLE_MOTOR_BW, ANGLE_MAX_SPEED);
  } else { // 정지
    digitalWrite(PIN_ANGLE_MOTOR_FW, ANGLE_MIN_SPEED);
    digitalWrite(PIN_ANGLE_MOTOR_BW, ANGLE_MIN_SPEED);
  }
}

void parseData() {
  if (isCheckOk()) {
    mode = readBuffer(2);
    life = readBuffer(3);
    int buzOctave = readBuffer(4);
    int buzNote = readBuffer(5);
    int motorState = readBuffer(6);
    int ledValue = readBuffer(7);
    shootMotorState = readBuffer(8);
    int d9Value = readBuffer(9);
    int d10Value = readBuffer(10);
    int angleMotorState = readBuffer(11);
    
    buzCtrl(buzOctave, buzNote);
    motorCtrl(motorState);
    ledCtrl(PIN_LED, ledValue);

    if (mode == 0) {

      controlShootMotorCtrl();

    } else {
      if (shootMotorState == 4) {
        ledCtrl(PIN_GEAR_MOTOR_FW, d9Value);
        ledCtrl(PIN_GEAR_MOTOR_BW, d10Value);
      } else {
      shootMotorCtrl(shootMotorState);
      }
    }
    angleMotorCtrl(angleMotorState);
  }
}

void getVibeValue() {
  writeValue(0xff);
  writeValue(0x7);
  writeValue(VIBE);
  int vibeValue = analogRead(PIN_VIBE);
  if (vibeValue > 600) {
    vibeState = 1;
  } else {
    vibeState = 0;
  }
  writeValue(vibeState);
  writeValue(life);
  int checkSum = 0;
  checkSum = (VIBE + vibeState + life) & 0xff;
  writeValue(checkSum);
  writeEnd();
}

void sendSensorValues() {
  getVibeValue();
}

void writeBuffer(int index, unsigned char c) {
  buf[index] = c;
}

void writeValue(unsigned char c) {
  Serial.write(c);
}

void writeEnd() {
  Serial.write(0xd);
  Serial.write(0xa);
}

void sendFloat(float value) {
  val.floatVal = value;
  writeValue(val.byteVal[0]);
  writeValue(val.byteVal[1]);
  writeValue(val.byteVal[2]);
  writeValue(val.byteVal[3]);
}

void bluetooth_read() {
  while (Serial.available()) {
    unsigned char c = Serial.read();
    Serial.println(c);
    updateDigitalPort(c);
  }
}

void setup() {
  initPorts();
  Serial.begin(115200);
  pixels.begin();
  pixels.clear();

  colorWipe(pixels.Color(40, 0, 0), 10);
  tone(PIN_BUZZER, 523, 80);
  colorWipe(pixels.Color(0, 40, 0), 10); 
  tone(PIN_BUZZER, 659, 80);
  colorWipe(pixels.Color(0, 0, 40), 10); 
  tone(PIN_BUZZER, 784, 80);  
//colorWipe(pixels.Color(0, 0, 0, 255), 30); 
  theaterChase(pixels.Color(40, 40, 40), 10); 
  tone(PIN_BUZZER, 1046, 1000);
  delay(1000);

  for(int i=0;i<4;i++){
   full_energy();
   delay(300);
   pixels.clear();
   pixels.show();
   delay(300);
  }
}

void loop() {
  bluetooth_read();

  if (mode == 1) {
    if(rotation > 1000) {
      rotation = 0;
      sendSensorValues();
    }
    rotation++;
  }
}

void callOK() {
  writeValue(0xff);
  writeValue(0xb);
  writeEnd();
}


void full_energy(){
  pixels.clear();    
  pixels.setPixelColor(9, pixels.Color(40, 0, 0));
  pixels.setPixelColor(10, pixels.Color(40, 0, 0));
  pixels.setPixelColor(11, pixels.Color(40, 0, 0));
  pixels.setPixelColor(12, pixels.Color(40, 0, 0));
  pixels.setPixelColor(13, pixels.Color(40, 0, 0));
  pixels.setPixelColor(14, pixels.Color(40, 0, 0));
  pixels.setPixelColor(15, pixels.Color(40, 0, 0));
  pixels.setPixelColor(16, pixels.Color(40, 0, 0));
  pixels.setPixelColor(17, pixels.Color(40, 0, 0));
  pixels.setPixelColor(18, pixels.Color(40, 0, 0));
  pixels.setPixelColor(19, pixels.Color(40, 0, 0));
  pixels.setPixelColor(20, pixels.Color(40, 0, 0));
  pixels.setPixelColor(21, pixels.Color(40, 0, 0));
  pixels.setPixelColor(22, pixels.Color(40, 0, 0));
  pixels.setPixelColor(23, pixels.Color(40, 0, 0));
  pixels.setPixelColor(24, pixels.Color(40, 0, 0));
  pixels.setPixelColor(25, pixels.Color(40, 0, 0));
  pixels.setPixelColor(26, pixels.Color(40, 0, 0));
  pixels.show();
 
}

void twostep_energy(){
  pixels.clear();
  pixels.setPixelColor(12, pixels.Color(40, 0, 0));
  pixels.setPixelColor(13, pixels.Color(40, 0, 0));
  pixels.setPixelColor(14, pixels.Color(40, 0, 0));
  pixels.setPixelColor(15, pixels.Color(40, 0, 0));
  pixels.setPixelColor(16, pixels.Color(40, 0, 0));
  pixels.setPixelColor(17, pixels.Color(40, 0, 0));
  pixels.setPixelColor(18, pixels.Color(40, 0, 0));
  pixels.setPixelColor(19, pixels.Color(40, 0, 0));
  pixels.setPixelColor(20, pixels.Color(40, 0, 0));
  pixels.setPixelColor(21, pixels.Color(40, 0, 0));
  pixels.setPixelColor(22, pixels.Color(40, 0, 0));
  pixels.setPixelColor(23, pixels.Color(40, 0, 0));
  pixels.show(); 
}

void onestep_energy(){
  pixels.clear();
  pixels.setPixelColor(15, pixels.Color(40, 0, 0));
  pixels.setPixelColor(16, pixels.Color(40, 0, 0));
  pixels.setPixelColor(17, pixels.Color(40, 0, 0));
  pixels.setPixelColor(18, pixels.Color(40, 0, 0));
  pixels.setPixelColor(19, pixels.Color(40, 0, 0));
  pixels.setPixelColor(20, pixels.Color(40, 0, 0));
  pixels.show(); 
}

void colorWipe(uint32_t c, uint8_t wait) {
  for(uint16_t i=0; i<pixels.numPixels(); i++) {
    pixels.setPixelColor(i, c);
    pixels.show();
    delay(wait);
  }
}

void theaterChase(uint32_t c, uint8_t wait) {
  for (int j=0; j<10; j++) {  
    for (int q=0; q < 3; q++) {
      for (uint16_t i=0; i < pixels.numPixels(); i=i+3) {
        pixels.setPixelColor(i+q, c);   
      }
      pixels.show();

      delay(wait);

      for (uint16_t i=0; i < pixels.numPixels(); i=i+3) {
        pixels.setPixelColor(i+q, 0);     
      }
    }
  }
}
